/*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.barista.api;

import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import fr.x9c.cadmium.kernel.Value;
import fr.x9c.cadmium.primitives.stdlib.Hash;
import fr.x9c.cadmium.support.values.ToValue;

/**
 * This class represents a Java access flag for a Java element.
 *
 * @author <a href="mailto:barista@x9c.fr">Xavier Clerc</a>
 * @version 1.0
 * @since 1.0
 */
public enum AccessFlag implements ToValue, Comparator<AccessFlag> {

    Public,
    Private,
    Protected,
    Static,
    Final,
    Super,
    Synchronized,
    Bridge,
    Volatile,
    Transient,
    Varargs,
    Native,
    Interface,
    Abstract,
    Strict,
    Synthetic,
    Annotation,
    Enum;

    /** Map from tag values to flags. */
    private static final Map<Integer, AccessFlag> MAP =
        new HashMap<Integer, AccessFlag>();

    /** Map initilization. */
    static {
        for (AccessFlag a : AccessFlag.values()) {
            MAP.put(a.toValue().asLong(), a);
        } // end for
    } // end static block

    /**
     * Tests whether the flag is a valid class flag.
     * @return <tt>true</tt> if the flag is a valid class flag,
     *         <tt>false</tt> otherwise
     */
    public boolean isClassFlag() {
        switch (this) {
        case Public:
        case Final:
        case Super:
        case Interface:
        case Abstract:
        case Synthetic:
        case Annotation:
        case Enum:
            return true;
        default:
            return false;
        } // end switch
    } // end method 'isClassFlag()'

    /**
     * Tests whether the flag is a valid inner-class flag.
     * @return <tt>true</tt> if the flag is a valid inner-class flag,
     *         <tt>false</tt> otherwise
     */
    public boolean isInnerClassFlag() {
        switch (this) {
        case Public:
        case Private:
        case Protected:
        case Static:
        case Final:
        case Super:
        case Interface:
        case Abstract:
        case Synthetic:
        case Annotation:
        case Enum:
            return true;
        default:
            return false;
        } // end switch
    } // end method 'isInnerClassFlag()'

    /**
     * Tests whether the flag is a valid field flag.
     * @return <tt>true</tt> if the flag is a valid field flag,
     *         <tt>false</tt> otherwise
     */
    public boolean isFieldFlag() {
        switch (this) {
        case Public:
        case Private:
        case Protected:
        case Static:
        case Final:
        case Volatile:
        case Transient:
        case Synthetic:
        case Enum:
            return true;
        default:
            return false;
        } // end switch
    } // end method 'isFieldFlag()'

    /**
     * Tests whether the flag is a valid method flag.
     * @return <tt>true</tt> if the flag is a valid method flag,
     *         <tt>false</tt> otherwise
     */
    public boolean isMethodFlag() {
        switch (this) {
        case Public:
        case Private:
        case Protected:
        case Static:
        case Final:
        case Synchronized:
        case Bridge:
        case Varargs:
        case Native:
        case Abstract:
        case Strict:
        case Synthetic:
            return true;
        default:
            return false;
        } // end switch
    } // end method 'isMethodFlag()'

    /**
     * Tests whether the flag is a valid constructor flag.
     * @return <tt>true</tt> if the flag is a valid constructor flag,
     *         <tt>false</tt> otherwise
     */
    public boolean isConstructorFlag() {
        switch (this) {
        case Public:
        case Private:
        case Protected:
        case Strict:
        case Varargs:
        case Synthetic:
            return true;
        default:
            return false;
        } // end switch
    } // end method 'isConstructorFlag()'

    /**
     * Returns the rank of a flag, used for comparison.
     * @return the rank of a flag, used for comparison
     */
    int getRank() {
        switch (this) {
        case Public:
        case Private:
        case Protected:
            return 1;
        case Abstract:
        case Static:
            return 2;
        case Final:
            return 3;
        case Synchronized:
        case Volatile:
        case Transient:
        case Native:
        case Strict:
            return 4;
        case Super:
        case Bridge:
        case Varargs:
        case Interface:
        case Synthetic:
        case Annotation:
        case Enum:
            return 5;
        default:
            assert false : "invalid flag";
            return 0;
        } // end switch
    } // end method 'getRank()'

    /**
     * {@inheritDoc}
     */
    public Value toValue() {
        return Hash.hashVariant(name());
    } // end method 'toValue()'

    /**
     * {@inheritDoc}
     */
    public int compare(final AccessFlag af1, final AccessFlag af2) {
        assert af1 != null : "null af1";
        assert af2 != null : "null af2";
        final int cmp = af1.getRank() - af2.getRank();
        if (cmp != 0) {
            return cmp;
        } else {
            return af1.name().compareTo(af2.name());
        } // end if/else
    } // end method 'toValue()'

    /**
     * Constructs an instance from a value.
     * @param v value to construct instance from - should not be <tt>null</tt>
     * @return the corresponding instance
     */
    static AccessFlag fromValue(final Value v) {
        assert v != null : "null v";
        return MAP.get(v.asLong());
    } // end method 'fromValue(Value)'

    /**
     * Tests whether a list contains only class flags.
     * @param l list to test - should not be <tt>null</tt>
     * @return <tt>true</tt> if the list contains only class flags,
     *         <tt>false</tt> otherwise
     */
    public static boolean areClassFlags(final List<AccessFlag> l) {
        for (AccessFlag a : l) {
            if (!a.isClassFlag()) {
                return false;
            } // end if
        } // end for
        return true;
    } // end method 'areClassFlags(List<AccessFlag>)'

    /**
     * Tests whether a list contains only inner-class flags.
     * @param l list to test - should not be <tt>null</tt>
     * @return <tt>true</tt> if the list contains only inner-class flags,
     *         <tt>false</tt> otherwise
     */
    public static boolean areInnerClassFlags(final List<AccessFlag> l) {
        for (AccessFlag a : l) {
            if (!a.isInnerClassFlag()) {
                return false;
            } // end if
        } // end for
        return true;
    } // end method 'areInnerClassFlags(List<AccessFlag>)'

    /**
     * Tests whether a list contains only field flags.
     * @param l list to test - should not be <tt>null</tt>
     * @return <tt>true</tt> if the list contains only field flags,
     *         <tt>false</tt> otherwise
     */
    public static boolean areFieldFlags(final List<AccessFlag> l) {
        for (AccessFlag a : l) {
            if (!a.isFieldFlag()) {
                return false;
            } // end if
        } // end for
        return true;
    } // end method 'areFieldFlags(List<AccessFlag>)'

    /**
     * Tests whether a list contains only method flags.
     * @param l list to test - should not be <tt>null</tt>
     * @return <tt>true</tt> if the list contains only method flags,
     *         <tt>false</tt> otherwise
     */
    public static boolean areMethodFlags(final List<AccessFlag> l) {
        for (AccessFlag a : l) {
            if (!a.isMethodFlag()) {
                return false;
            } // end if
        } // end for
        return true;
    } // end method 'areMethodFlags(List<AccessFlag>)'

    /**
     * Tests whether a list contains only constructor flags.
     * @param l list to test - should not be <tt>null</tt>
     * @return <tt>true</tt> if the list contains only constructor flags,
     *         <tt>false</tt> otherwise
     */
    public static boolean areConstructorFlags(final List<AccessFlag> l) {
        for (AccessFlag a : l) {
            if (!a.isConstructorFlag()) {
                return false;
            } // end if
        } // end for
        return true;
    } // end method 'areConstructorFlags(List<AccessFlag>)'

    /**
     * Converts a flag list into a string.
     * @param l list to convert - should not be <tt>null</tt>
     * @return the concatenation of all list elements
     */
    public static String listToString(final List<AccessFlag> l) {
        assert l != null : "null l";
        final StringBuilder sb = new StringBuilder();
        final List<AccessFlag> l2 = new LinkedList<AccessFlag>(l);
        Collections.sort(l2);
        boolean first = true;
        for (AccessFlag a : l2) {
            if (first) {
                first = false;
            } else {
                sb.append(" ");
            } // end if/else
            sb.append(a.toString());
        } // end for
        return sb.toString();
    } // end method 'listToString(List<AccessFlag>)'

} // end class 'AccessFlag'
